import * as interfaces from '../interfaces/index.js';
import * as plugins from '../plugins.js';
import { WorkerEvent } from '../classes.workerevent.js';
import { kvHandlerInstance } from '../classes.kvhandler.js';

declare const fetch: plugins.cloudflareTypes.Fetcher['fetch']; 

declare var caches: any;
export const cacheResponder: interfaces.TRequestResponser = async (cworkerEventArg: WorkerEvent) => {
  const host = cworkerEventArg.request.headers.get('Host');
  const appHashKey = `${host.toLowerCase()}_appHash`;
  const appHash = await kvHandlerInstance.getFromKv(appHashKey);

  const cache = caches.default;
  let response: Response = await cache.match(cworkerEventArg.request);

  if (
    response &&
    response.headers.get('appHash') &&
    response.headers.get('appHash') !== appHash
  ) {
    response = null;
  }

  if (response) {
    cworkerEventArg.setResponse(response);
  } else {
    response = await handleNewRequest(cworkerEventArg.request);
    if (response) {
      cworkerEventArg.addToWaitList(new Promise<void>(async (resolve, reject) => {
        const newAppHash = response.headers.get('appHash');
        if (newAppHash) {
          await kvHandlerInstance.putInKv(appHashKey, newAppHash);
        }
        resolve();
      }));
      cworkerEventArg.addToWaitList(buildCacheResponse(cache, cworkerEventArg.request, response));
      cworkerEventArg.setResponse(response);
    }
  }
};

/**
 * @param {Request} originalRequest
 */
const handleNewRequest = async (originalRequest: plugins.cloudflareTypes.Request): Promise<Response> => {
  console.log('answering from origin');
  const originResponse: any = await fetch(
    originalRequest
  );

  // lets capture status
  if (originResponse.status > 399) {
    return null;
  }

  const responseClientPassThroughStream = new TransformStream();
  originResponse.body.pipeTo(responseClientPassThroughStream.writable);

  // build response for client
  const clientHeaders = new Headers();
  for (const kv of originResponse.headers.entries()) {
    clientHeaders.append(kv[0], kv[1]);
  }
  clientHeaders.append('SERVEZONE_ROUTE', 'LOSSLESS_EDGE_ORIGIN_INITIAL');
  const responseForClient = new Response(responseClientPassThroughStream.readable, {
    ...originResponse,
    headers: clientHeaders
  });

  // lets return the responses
  return responseForClient;
};

const buildCacheResponse = async (cache, matchRequest: plugins.cloudflareTypes.Request, originResponse: any) => {
  const cacheHeaders = new Headers();
  for (const kv of originResponse.headers.entries()) {
    cacheHeaders.append(kv[0], kv[1]);
  }
  cacheHeaders.delete('SERVEZONE_ROUTE');
  cacheHeaders.append('SERVEZONE_ROUTE', 'LOSSLESS_EDGE_CACHE');
  cacheHeaders.delete('Cache-Control');
  cacheHeaders.append('Cache-Control', 'public, max-age=60');
  cacheHeaders.delete('Expires');
  cacheHeaders.append('Expires', new Date(Date.now() + 60 * 1000).toUTCString());

  const responseForCache = new Response(await originResponse.clone().body, {
    ...originResponse,
    headers: cacheHeaders
  });
  await cache.put(matchRequest, responseForCache);
};
